@@ -1,5 +1,3 @@
-#!/usr/bin/python
-#
 # Urwid MonitoredList class
 #    Copyright (C) 2004-2011  Ian Ward
 #
@@ -17,25 +15,39 @@
 #    License along with this library; if not, write to the Free Software
 #    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 #
-# Urwid web site: http://excess.org/urwid/
+# Urwid web site: https://urwid.org/
 
-from __future__ import division, print_function
+from __future__ import annotations
 
-from urwid.compat import PYTHON3, xrange
+import functools
+import typing
+import warnings
+
+if typing.TYPE_CHECKING:
+    from collections.abc import Callable
+
+    from typing_extensions import ParamSpec
+
+    ArgSpec = ParamSpec("ArgSpec")
+    Ret = typing.TypeVar("Ret")
 
 
-def _call_modified(fn):
-    def call_modified_wrapper(self, *args, **kwargs):
+def _call_modified(fn: Callable[ArgSpec, Ret]) -> Callable[ArgSpec, Ret]:
+    @functools.wraps(fn)
+    def call_modified_wrapper(self: MonitoredList, *args, **kwargs):
         rval = fn(self, *args, **kwargs)
         self._modified()
         return rval
+
     return call_modified_wrapper
 
+
 class MonitoredList(list):
     """
     This class can trigger a callback any time its contents are changed
     with the usual list operations append, extend, etc.
     """
+
     def _modified(self):
         pass
 
@@ -63,18 +75,16 @@ class MonitoredList(list):
         self._modified = callback
 
     def __repr__(self):
-        return "%s(%r)" % (self.__class__.__name__, list(self))
+        return f"{self.__class__.__name__}({list(self)!r})"
 
     __add__ = _call_modified(list.__add__)
     __delitem__ = _call_modified(list.__delitem__)
-    if not PYTHON3:
-        __delslice__ = _call_modified(list.__delslice__)
+
     __iadd__ = _call_modified(list.__iadd__)
     __imul__ = _call_modified(list.__imul__)
     __rmul__ = _call_modified(list.__rmul__)
     __setitem__ = _call_modified(list.__setitem__)
-    if not PYTHON3:
-        __setslice__ = _call_modified(list.__setslice__)
+
     append = _call_modified(list.append)
     extend = _call_modified(list.extend)
     insert = _call_modified(list.insert)
@@ -82,7 +92,7 @@ class MonitoredList(list):
     remove = _call_modified(list.remove)
     reverse = _call_modified(list.reverse)
     sort = _call_modified(list.sort)
-    if hasattr(list, 'clear'):
+    if hasattr(list, "clear"):
         clear = _call_modified(list.clear)
 
 
@@ -91,6 +101,7 @@ class MonitoredFocusList(MonitoredList):
     This class can trigger a callback any time its contents are modified,
     before and/or after modification, and any time the focus index is changed.
     """
+
     def __init__(self, *argl, **argd):
         """
         This is a list that tracks one item as the focus item.  If items
@@ -112,31 +123,36 @@ class MonitoredFocusList(MonitoredList):
         >>> ml
         MonitoredFocusList([], focus=None)
         """
-        focus = argd.pop('focus', 0)
+        focus = argd.pop("focus", 0)
 
-        super(MonitoredFocusList, self).__init__(*argl, **argd)
+        super().__init__(*argl, **argd)
 
         self._focus = focus
         self._focus_modified = lambda ml, indices, new_items: None
 
     def __repr__(self):
-        return "%s(%r, focus=%r)" % (
-            self.__class__.__name__, list(self), self.focus)
+        return f"{self.__class__.__name__}({list(self)!r}, focus={self.focus!r})"
 
-    def _get_focus(self):
+    @property
+    def focus(self) -> int | None:
         """
+        Get/set the focus index.  This value is read as None when the list
+        is empty, and may only be set to a value between 0 and len(self)-1
+        or an IndexError will be raised.
+
         Return the index of the item "in focus" or None if
         the list is empty.
 
-        >>> MonitoredFocusList([1,2,3], focus=2)._get_focus()
+        >>> MonitoredFocusList([1,2,3], focus=2).focus
         2
-        >>> MonitoredFocusList()._get_focus()
+        >>> MonitoredFocusList().focus
         """
         if not self:
             return None
         return self._focus
 
-    def _set_focus(self, index):
+    @focus.setter
+    def focus(self, index: int) -> None:
         """
         index -- index into this list, any index out of range will
             raise an IndexError, except when the list is empty and
@@ -149,11 +165,11 @@ class MonitoredFocusList(MonitoredList):
         instance with set_focus_changed_callback().
 
         >>> ml = MonitoredFocusList([9, 10, 11])
-        >>> ml._set_focus(2); ml._get_focus()
+        >>> ml.focus = 2; ml.focus
         2
-        >>> ml._set_focus(0); ml._get_focus()
+        >>> ml.focus = 0; ml.focus
         0
-        >>> ml._set_focus(-2)
+        >>> ml.focus = -2
         Traceback (most recent call last):
         ...
         IndexError: focus index is out of range: -2
@@ -162,24 +178,36 @@ class MonitoredFocusList(MonitoredList):
             self._focus = 0
             return
         if index < 0 or index >= len(self):
-            raise IndexError('focus index is out of range: %s' % (index,))
+            raise IndexError(f"focus index is out of range: {index}")
         if index != int(index):
-            raise IndexError('invalid focus index: %s' % (index,))
+            raise IndexError(f"invalid focus index: {index}")
         index = int(index)
         if index != self._focus:
             self._focus_changed(index)
         self._focus = index
 
-    focus = property(_get_focus, _set_focus, doc="""
-        Get/set the focus index.  This value is read as None when the list
-        is empty, and may only be set to a value between 0 and len(self)-1
-        or an IndexError will be raised.
-        """)
+    def _get_focus(self) -> int | None:
+        warnings.warn(
+            f"method `{self.__class__.__name__}._get_focus` is deprecated, "
+            f"please use `{self.__class__.__name__}.focus` property",
+            DeprecationWarning,
+            stacklevel=3,
+        )
+        return self.focus
 
-    def _focus_changed(self, new_focus):
+    def _set_focus(self, index: int) -> None:
+        warnings.warn(
+            f"method `{self.__class__.__name__}._set_focus` is deprecated, "
+            f"please use `{self.__class__.__name__}.focus` property",
+            DeprecationWarning,
+            stacklevel=3,
+        )
+        self.focus = index
+
+    def _focus_changed(self, new_focus: int):
         pass
 
-    def set_focus_changed_callback(self, callback):
+    def set_focus_changed_callback(self, callback: Callable[[int], typing.Any]) -> None:
         """
         Assign a callback to be called when the focus index changes
         for any reason.  The callback is in the form:
@@ -229,7 +257,7 @@ class MonitoredFocusList(MonitoredList):
         """
         self._validate_contents_modified = callback
 
-    def _adjust_focus_on_contents_modified(self, slc, new_items=()):
+    def _adjust_focus_on_contents_modified(self, slc: slice, new_items=()) -> int:
         """
         Default behaviour is to move the focus to the item following
         any removed items, unless that item was simply replaced.
@@ -240,7 +268,7 @@ class MonitoredFocusList(MonitoredList):
         """
         num_new_items = len(new_items)
         start, stop, step = indices = slc.indices(len(self))
-        num_removed = len(list(xrange(*indices)))
+        num_removed = len(list(range(*indices)))
 
         focus = self._validate_contents_modified(indices, new_items)
         if focus is not None:
@@ -254,20 +282,20 @@ class MonitoredFocusList(MonitoredList):
             if stop <= focus:
                 focus += num_new_items - (stop - start)
 
-        else:
+        else:  # noqa: PLR5501
             if not num_new_items:
                 # extended slice being removed
-                if focus in xrange(start, stop, step):
+                if focus in range(start, stop, step):
                     focus += 1
 
                 # adjust for removed items
-                focus -= len(list(xrange(start, min(focus, stop), step)))
+                focus -= len(list(range(start, min(focus, stop), step)))
 
-        return min(focus, len(self) + num_new_items - num_removed -1)
+        return min(focus, len(self) + num_new_items - num_removed - 1)
 
     # override all the list methods that modify the list
 
-    def __delitem__(self, y):
+    def __delitem__(self, y: int | slice):
         """
         >>> ml = MonitoredFocusList([0,1,2,3,4], focus=2)
         >>> del ml[3]; ml
@@ -296,13 +324,12 @@ class MonitoredFocusList(MonitoredList):
         if isinstance(y, slice):
             focus = self._adjust_focus_on_contents_modified(y)
         else:
-            focus = self._adjust_focus_on_contents_modified(slice(y,
-                y+1 or None))
-        rval = super(MonitoredFocusList, self).__delitem__(y)
-        self._set_focus(focus)
+            focus = self._adjust_focus_on_contents_modified(slice(y, y + 1 or None))
+        rval = super().__delitem__(y)
+        self.focus = focus
         return rval
 
-    def __setitem__(self, i, y):
+    def __setitem__(self, i: int | slice, y):
         """
         >>> def modified(indices, new_items):
         ...     print("range%r <- %r" % (indices, new_items))
@@ -334,19 +361,12 @@ class MonitoredFocusList(MonitoredList):
         if isinstance(i, slice):
             focus = self._adjust_focus_on_contents_modified(i, y)
         else:
-            focus = self._adjust_focus_on_contents_modified(slice(i, i+1 or None), [y])
-        rval = super(MonitoredFocusList, self).__setitem__(i, y)
-        self._set_focus(focus)
+            focus = self._adjust_focus_on_contents_modified(slice(i, i + 1 or None), [y])
+        rval = super().__setitem__(i, y)
+        self.focus = focus
         return rval
 
-    if not PYTHON3:
-        def __delslice__(self, i, j):
-            return self.__delitem__(slice(i,j))
-
-        def __setslice__(self, i, j, y):
-            return self.__setitem__(slice(i, j), y)
-
-    def __imul__(self, n):
+    def __imul__(self, n: int):
         """
         >>> def modified(indices, new_items):
         ...     print("range%r <- %r" % (indices, list(new_items)))
@@ -362,12 +382,11 @@ class MonitoredFocusList(MonitoredList):
         None
         """
         if n > 0:
-            focus = self._adjust_focus_on_contents_modified(
-                slice(len(self), len(self)), list(self)*(n-1))
-        else: # all contents are being removed
+            focus = self._adjust_focus_on_contents_modified(slice(len(self), len(self)), list(self) * (n - 1))
+        else:  # all contents are being removed
             focus = self._adjust_focus_on_contents_modified(slice(0, len(self)))
-        rval = super(MonitoredFocusList, self).__imul__(n)
-        self._set_focus(focus)
+        rval = super().__imul__(n)
+        self.focus = focus
         return rval
 
     def append(self, item):
@@ -379,10 +398,9 @@ class MonitoredFocusList(MonitoredList):
         >>> ml.append(6)
         range(3, 3, 1) <- [6]
         """
-        focus = self._adjust_focus_on_contents_modified(
-            slice(len(self), len(self)), [item])
-        rval = super(MonitoredFocusList, self).append(item)
-        self._set_focus(focus)
+        focus = self._adjust_focus_on_contents_modified(slice(len(self), len(self)), [item])
+        rval = super().append(item)
+        self.focus = focus
         return rval
 
     def extend(self, items):
@@ -394,13 +412,12 @@ class MonitoredFocusList(MonitoredList):
         >>> ml.extend((6,7,8))
         range(3, 3, 1) <- [6, 7, 8]
         """
-        focus = self._adjust_focus_on_contents_modified(
-            slice(len(self), len(self)), items)
-        rval = super(MonitoredFocusList, self).extend(items)
-        self._set_focus(focus)
+        focus = self._adjust_focus_on_contents_modified(slice(len(self), len(self)), items)
+        rval = super().extend(items)
+        self.focus = focus
         return rval
 
-    def insert(self, index, item):
+    def insert(self, index: int, item):
         """
         >>> ml = MonitoredFocusList([0,1,2,3], focus=2)
         >>> ml.insert(-1, -1); ml
@@ -410,13 +427,12 @@ class MonitoredFocusList(MonitoredList):
         >>> ml.insert(3, -3); ml
         MonitoredFocusList([-2, 0, 1, -3, 2, -1, 3], focus=4)
         """
-        focus = self._adjust_focus_on_contents_modified(slice(index, index),
-            [item])
-        rval = super(MonitoredFocusList, self).insert(index, item)
-        self._set_focus(focus)
+        focus = self._adjust_focus_on_contents_modified(slice(index, index), [item])
+        rval = super().insert(index, item)
+        self.focus = focus
         return rval
 
-    def pop(self, index=-1):
+    def pop(self, index: int = -1):
         """
         >>> ml = MonitoredFocusList([-2,0,1,-3,2,3], focus=4)
         >>> ml.pop(3); ml
@@ -432,10 +448,9 @@ class MonitoredFocusList(MonitoredList):
         2
         MonitoredFocusList([0, 1], focus=1)
         """
-        focus = self._adjust_focus_on_contents_modified(slice(index,
-            index+1 or None))
-        rval = super(MonitoredFocusList, self).pop(index)
-        self._set_focus(focus)
+        focus = self._adjust_focus_on_contents_modified(slice(index, index + 1 or None))
+        rval = super().pop(index)
+        self.focus = focus
         return rval
 
     def remove(self, value):
@@ -449,10 +464,9 @@ class MonitoredFocusList(MonitoredList):
         MonitoredFocusList([0, 1, 2, -1], focus=2)
         """
         index = self.index(value)
-        focus = self._adjust_focus_on_contents_modified(slice(index,
-            index+1 or None))
-        rval = super(MonitoredFocusList, self).remove(value)
-        self._set_focus(focus)
+        focus = self._adjust_focus_on_contents_modified(slice(index, index + 1 or None))
+        rval = super().remove(value)
+        self.focus = focus
         return rval
 
     def reverse(self):
@@ -461,8 +475,8 @@ class MonitoredFocusList(MonitoredList):
         >>> ml.reverse(); ml
         MonitoredFocusList([4, 3, 2, 1, 0], focus=3)
         """
-        rval = super(MonitoredFocusList, self).reverse()
-        self._set_focus(max(0, len(self) - self._focus - 1))
+        rval = super().reverse()
+        self.focus = max(0, len(self) - self._focus - 1)
         return rval
 
     def sort(self, **kwargs):
@@ -472,27 +486,26 @@ class MonitoredFocusList(MonitoredList):
         MonitoredFocusList([-3, -2, -1, 0, 1, 2, 3], focus=5)
         """
         if not self:
-            return
+            return None
         value = self[self._focus]
-        rval = super(MonitoredFocusList, self).sort(**kwargs)
-        self._set_focus(self.index(value))
+        rval = super().sort(**kwargs)
+        self.focus = self.index(value)
         return rval
 
-    if hasattr(list, 'clear'):
+    if hasattr(list, "clear"):
+
         def clear(self):
             focus = self._adjust_focus_on_contents_modified(slice(0, 0))
-            rval = super(MonitoredFocusList, self).clear()
-            self._set_focus(focus)
+            rval = super().clear()
+            self.focus = focus
             return rval
 
 
-
-
-
 def _test():
     import doctest
+
     doctest.testmod()
 
-if __name__=='__main__':
-    _test()
 
+if __name__ == "__main__":
+    _test()
